@@ -32,6 +32,7 @@
 
 
 #undef CONFIG_ATIME_FILE
+#undef CONFIG_SKETCH_FILE
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -41,8 +42,10 @@
 #include <strings.h>
 #include <stdarg.h>
 #include <sys/ioctl.h>
+#include <sys/wait.h>
 #include <uuid/uuid.h>
 #include <string.h>
+#include <errno.h>
 #include "mkfs.h"
 
 
@@ -58,7 +61,11 @@ typedef __u64  blocknr_t;
  */
 #define MOUNTS			"/etc/mtab"
 #define LINE_BUFFER_SIZE	256  /* Line buffer size for reading mtab */
-#define COMMAND_BUFFER_SIZE     1024
+#define BLOCKSIZE_BUFFER_SIZE	100
+
+#define BADBLOCKS_NAME		"badblocks"
+
+static const char badblocks[] = "/sbin/" BADBLOCKS_NAME;
 
 /*
  * Command interface primitives
@@ -126,7 +133,7 @@ static const unsigned nr_initial_inodes = 3;  /* root directory + .sketch +
 static const __u64 first_cno = 1; /* Number of the first checkpoint */
 
 /* Segment layout information (per partial segment) */
-#define MAX_FILES        10
+#define MAX_FILES        16
 
 struct nilfs_file_info {
 	ino_t ino;
@@ -180,7 +187,8 @@ struct nilfs_segment_ref {
 	__u64 cno;                /* checkpoint number */
 };
 
-static void init_disk_layout(struct nilfs_disk_info *, int, const char *, struct mkfs_options *);
+static void init_disk_layout(struct nilfs_disk_info *, int, const char *,
+			     struct mkfs_options *);
 
 
 static inline blocknr_t count_free_blocks(struct nilfs_disk_info *di)
@@ -244,6 +252,7 @@ static void commit_segment(void);
 static void make_rootdir(void);
 static void make_sketch(void);
 static void make_dot_nilfs(void);
+static void make_reserved_files(void);
 
 
 static inline struct nilfs_segment_ref *get_last_segment(void)
@@ -423,7 +432,8 @@ init_disk_layout(struct nilfs_disk_info *di, int fd, const char *device,
 		too_small_segment(di->blocks_per_segment,
 				  first_segblk + NILFS_PSEG_MIN_BLOCKS);
 
-	di->nsegments = (dev_size >> di->blkbits) / di->blocks_per_segment;
+	di->nsegments = (NILFS_SB2_OFFSET_BYTES(dev_size) >> di->blkbits) /
+		di->blocks_per_segment;
 	min_nsegments = nilfs_min_nsegments(di, opts->r_segments_percentage);
 	if (di->nsegments < min_nsegments)
 		perr("Error: too small device."
@@ -511,6 +521,7 @@ int main(int argc, char *argv[])
 	struct nilfs_disk_info diskinfo, *di = &diskinfo;
 	struct mkfs_options *opts = &options;
 	struct nilfs_segment_info *si;
+	struct stat statbuf;
 	const char *device;
 	int fd;
 
@@ -519,6 +530,11 @@ int main(int argc, char *argv[])
 
 	blocksize = opts->blocksize;
 
+	if (stat(device, &statbuf) != 0)
+		perr("Error: cannot find %s: %s", device, strerror(errno));
+	else if (!S_ISREG(statbuf.st_mode) && !S_ISBLK(statbuf.st_mode))
+		perr("Error: device must be a block device or a file");
+
 	if (opts->cflag)
 		disk_scan(device, opts);  /* check the block device */
 	if ((fd = open(device, O_RDWR)) < 0)
@@ -531,6 +547,10 @@ int main(int argc, char *argv[])
 	add_file(si, NILFS_ROOT_INO, 1, 0);
 	add_file(si, NILFS_SKETCH_INO, 0, 0);
 	add_file(si, NILFS_NILFS_INO, 0, 0);
+	add_file(si, NILFS_ATIME_INO, 0, 0);
+	add_file(si, 1, 0, 0);
+	add_file(si, 8, 0, 0);
+	add_file(si, 9, 0, 0);
 	add_file(si, NILFS_IFILE_INO, count_ifile_blocks(blocksize), 0);
 	add_file(si, NILFS_CPFILE_INO, count_cpfile_blocks(blocksize), 0);
 	add_file(si, NILFS_SUFILE_INO, count_sufile_blocks(blocksize), 0);
@@ -550,6 +570,7 @@ int main(int argc, char *argv[])
 	make_sketch();   /* Make sketch file */
 	make_dot_nilfs(); /* Make .nilfs */
 	make_rootdir();  /* Make root directory */
+	make_reserved_files();
 	commit_segment();
 
 	commit_super_block(di, get_last_segment());
@@ -565,16 +586,56 @@ int main(int argc, char *argv[])
  */
 static void disk_scan(const char *device, struct mkfs_options *opts)
 {
-	char buf[COMMAND_BUFFER_SIZE];
+	struct stat statbuf;
+	pid_t pid;
+	int status;
+
+	if (stat(badblocks, &statbuf) != 0) {
+		pinfo("Warning: %s not found.", badblocks);
+		return;
+	}
 
-	if (snprintf(buf, COMMAND_BUFFER_SIZE, "badblocks -b %ld %s %s %s\n",
-		     opts->blocksize, opts->quiet ? "" : "-s",
-		     (opts->cflag > 1) ? "-w" : "",
-		     device) < COMMAND_BUFFER_SIZE)
-		perr("Internal error: command line buffer overflow");
 	if (!opts->quiet)
 		pinfo("checking blocks");
-	system(buf);
+
+	pid = fork();
+	if (pid == 0) {
+		char bszbuf[BLOCKSIZE_BUFFER_SIZE];
+		const char *args[7];
+		ssize_t n;
+		int i = 0;
+
+		if (setgid(getgid()) < 0)
+			perr("Error: failed to drop setgid privileges");
+		if (setuid(getuid()) < 0)
+			perr("Error: failed to drop setuid privileges");
+		args[i++] = badblocks;
+		args[i++] = "-b";
+		n = snprintf(bszbuf, BLOCKSIZE_BUFFER_SIZE, "%ld",
+			     opts->blocksize);
+		if (n < 0 || n >= BLOCKSIZE_BUFFER_SIZE)
+			perr("Internal error: blocksize buffer overflow");
+		args[i++] = bszbuf;
+
+		if (!opts->quiet)
+			args[i++] = "-s";
+		if (opts->cflag > 1)
+			args[i++] = "-w";
+		args[i++] = device;
+		args[i] = NULL;
+		execv(badblocks, (char **)args);
+		exit(1); /* reach only if failed */
+	} else if (pid != -1) {
+		if (wait(&status) < 0)
+			perr("Error: cannot wait child");
+		else if (!WIFEXITED(status))
+			perr("Error: check aborted");
+		else if (WEXITSTATUS(status))
+			perr("Error: check failed with status(%d)",
+			     WEXITSTATUS(status));
+	} else {
+		perr("Error: cannot fork: %s", strerror(errno));
+	}
 }
 
 static void check_mount(int fd, const char *device)
@@ -654,6 +715,60 @@ static void read_disk_header(int fd, const char *device)
 	}
 }
 
+#define MAX_NBLOCKS_CLEAR_BUFFER	8
+
+static int erase_disk_range(int fd, off64_t offset, size_t count)
+{
+	void *buffer;
+	size_t size, bufsz;
+	int ret = -1;
+
+	for (bufsz = blocksize * MAX_NBLOCKS_CLEAR_BUFFER;
+	     bufsz >= blocksize; bufsz >>= 1) {
+		buffer = malloc(bufsz);
+		if (buffer != NULL)
+			break;
+	}
+	if (bufsz < blocksize)
+		cannot_allocate_memory();
+
+	memset(buffer, 0, bufsz);
+
+	if (lseek64(fd, offset, SEEK_SET) < 0)
+		goto failed;
+
+	while (count > 0) {
+		size = count > bufsz ? bufsz : count;
+		if (write(fd, buffer, size) < size)
+			goto failed;
+		count -= size;
+	}
+	ret = 0;
+
+ failed:
+	free(buffer);
+	return ret;
+}
+
+static int erase_disk(int fd, struct nilfs_disk_info *di)
+{
+	int ret;
+
+	BUG_ON(di->dev_size < NILFS_DISK_ERASE_SIZE ||
+	       di->dev_size - NILFS_DISK_ERASE_SIZE < NILFS_SB_OFFSET_BYTES);
+
+	/* Erase tail of partition */
+	ret = erase_disk_range(fd, di->dev_size - NILFS_DISK_ERASE_SIZE,
+			       NILFS_DISK_ERASE_SIZE);
+	if (ret == 0) {
+		/* Erase head of partition */
+		ret = erase_disk_range(fd, NILFS_SB_OFFSET_BYTES,
+				       NILFS_DISK_ERASE_SIZE -
+				       NILFS_SB_OFFSET_BYTES);
+	}
+	return ret;
+}
+
 static void write_disk(int fd, struct nilfs_disk_info *di,
 		       struct mkfs_options *opts)
 {
@@ -668,6 +783,9 @@ static void write_disk(int fd, struct nilfs_disk_info *di,
 		      blocksize, di->device, di->dev_size);
 	}
 	if (!opts->nflag) {
+		if (erase_disk(fd, di) < 0)
+			goto failed_to_write;
+
 		/* Writing segments */
 		for (i = 0, si = di->seginfo; i < di->nseginfo; i++, si++) {
 			lseek64(fd, si->start * blocksize, SEEK_SET);
@@ -681,11 +799,18 @@ static void write_disk(int fd, struct nilfs_disk_info *di,
 		if (fsync(fd) < 0)
 			goto failed_to_write;
 
-		/* Writing super block */
-		blocknr = NILFS_SB_OFFSET_BYTES / blocksize;
-		lseek64(fd, blocknr * blocksize, SEEK_SET);
-		if (write(fd, map_disk_buffer(blocknr, 1), blocksize) < 0 ||
-		    fsync(fd) < 0)
+		/* Writing primary super block */
+		if (lseek64(fd, NILFS_SB_OFFSET_BYTES, SEEK_SET) < 0 ||
+		    write(fd, raw_sb, sizeof(*raw_sb)) < 0)
+			goto failed_to_write;
+
+		/* Writing secondary super block */
+		if (lseek64(fd, NILFS_SB2_OFFSET_BYTES(di->dev_size),
+			    SEEK_SET) < 0 ||
+		    write(fd, raw_sb, sizeof(*raw_sb)) < 0)
+			goto failed_to_write;
+
+		if (fsync(fd) < 0)
 			goto failed_to_write;
 	}
 	if (!opts->quiet)
@@ -745,7 +870,7 @@ static void parse_options(int argc, char *argv[], struct mkfs_options *opts)
 {
 	int c, show_version_only = 0;
 
-	while ((c = getopt(argc, argv, "b:B:cL:m:nq:VP:")) != EOF) {
+	while ((c = getopt(argc, argv, "b:B:cL:m:nqVP:")) != EOF) {
 		switch (c) {
 		case 'b':
 			opts->blocksize = atol(optarg);
@@ -988,6 +1113,7 @@ static void make_empty_dir(ino_t dir_ino, ino_t parent_ino)
 	memcpy(de->name, "..\0\0\0\0\0", 8);
 //	rec_len += de->rec_len;
 
+#if defined(CONFIG_SKETCH_FILE)
 	de = next_dir_entry(de);
 	de->inode = cpu_to_le64(NILFS_SKETCH_INO);
 	de->name_len = 7;
@@ -996,6 +1122,7 @@ static void make_empty_dir(ino_t dir_ino, ino_t parent_ino)
 	de->file_type = NILFS_FT_REG_FILE;
 	memcpy(de->name, ".sketch", 8);
 //	rec_len += de->rec_len;
+#endif
 
 	de = next_dir_entry(de);
 	de->inode = cpu_to_le64(NILFS_NILFS_INO);
@@ -1024,6 +1151,14 @@ static void make_dot_nilfs(void)
 	init_inode(NILFS_NILFS_INO, DT_REG, 0644, 0);
 }
 
+static void make_reserved_files(void)
+{
+	init_inode(NILFS_ATIME_INO, DT_REG, 0, 0);
+	init_inode(1, DT_REG, 0, 0);
+	init_inode(8, DT_REG, 0, 0);
+	init_inode(9, DT_REG, 0, 0);
+}
+
 static void *
 map_segsum_info(blocknr_t start, unsigned long *offset, unsigned item_size)
 {
@@ -1084,21 +1219,21 @@ static void update_blocknr(struct nilfs_file_info *fi,
 
 static void prepare_blockgrouped_file(blocknr_t blocknr)
 {
-	struct nilfs_persistent_group_desc *desc;
+	struct nilfs_palloc_group_desc *desc;
 	const unsigned group_descs_per_block
-		= blocksize / sizeof(struct nilfs_persistent_group_desc);
+		= blocksize / sizeof(struct nilfs_palloc_group_desc);
 	int i;
 
 	for (i = 0, desc = map_disk_buffer(blocknr, 1);
 	     i < group_descs_per_block; i++, desc++)
-		desc->pg_nfrees = cpu_to_le32(blocksize * NILFS_CHAR_BIT);
+		desc->pg_nfrees = cpu_to_le32(blocksize * 8 /* CHAR_BIT */);
 	map_disk_buffer(blocknr + 1, 1); /* Initialize bitmap block */
 }
 
 static inline void
 alloc_blockgrouped_file_entry(blocknr_t blocknr, unsigned long nr)
 {
-	struct nilfs_persistent_group_desc *desc = map_disk_buffer(blocknr, 1);
+	struct nilfs_palloc_group_desc *desc = map_disk_buffer(blocknr, 1);
 					/* allways use the first group */
 	void *bitmap = map_disk_buffer(blocknr + 1, 1);
 
@@ -1494,8 +1629,3 @@ static void commit_super_block(struct nilfs_disk_info *di,
 			    NILFS_SB_BYTES);
 	raw_sb->s_sum = cpu_to_le32(sbsum);
 }
-
-
-/* Local Variables:		*/
-/* eval: (c-set-style "linux")	*/
-/* End:				*/
